"""A translation of a Scheme interpreter in Python, modeling the
metacircular evaluator in Chapter 4 of The Structure and
Interpretation of Computer Programs.

Danny Yoo (dyoo@hkn.eecs.berkeley.edu)

One major deficiency of this implementation is that it doesn't support
tail recursion yet.  I'll have to see if such a thing is possible in
standard Python.

This source code is for the public domain; I'm writing it just for the
fun of it.  This is certainly not meant to be useful.  *grin*

"""

from __future__ import nested_scopes

import types, math, sys, pprint
from symbol import Symbol, true, false, isSymbol


"""Let's define a personalized SchemeError exception that will be
thrown if bad things happen."""

class SchemeError(Exception): pass

evalInUnderlyingPython = eval
applyInUnderlyingPython = apply


def eval(exp, env):
    """Evaluates an expression 'exp' in an environment 'env'.

    Exercise 4.3 asks us to rewrite this in a more natural
    data-directed manner.
    """
    if isSelfEvaluating(exp):
        return exp
    if isVariable(exp):
        return lookupVariableValue(exp, env)
    if isQuoted(exp):
        return textOfQuotation(exp)
    if isAssignment(exp):
        return evalAssignment(exp, env)
    if isDefinition(exp):
        return evalDefinition(exp, env)
    if isIf(exp):
        return evalIf(exp, env)
    if isLambda(exp):
        return makeProcedure(lambdaParameters(exp), lambdaBody(exp),
                             env)
    if isBegin(exp):
        return evalSequence(beginActions(exp), env)
    if isCond(exp):
        return eval(condToIf(exp), env)
    if isApplication(exp):
        return apply(eval(operator(exp), env),
                     listOfValues(operands(exp), env))
    raise SchemeError, "Unknown expression type -- eval " + exp


def apply(procedure, arguments):
    """Applies a procedure on a list of arguments."""
    if isPrimitiveProcedure(procedure):
        return applyPrimitiveProcedure(procedure, arguments)
    if isCompoundProcedure(procedure):
        return evalSequence(procedureBody(procedure),
                            extendEnvironment(procedureParameters(procedure),
                                              arguments,
                                              procedureEnvironment(procedure)))
    raise SchemeError, "Unknown procedure type -- apply " + str(procedure)



def listOfValues(exps, env):
    """Given a list of expressions, returns a new list containing the
    values of evaluating each on of them."""
    if isNoOperands(exps): return []
    return [eval(firstOperand(exps), env)] + \
           listOfValues(restOperands(exps), env)

def evalIf(exp, env):
    if isTrue(eval(ifPredicate(exp), env)):
        return eval(ifConsequent(exp), env)
    return eval(ifAlternative(exp), env)


def evalSequence(exps, env):
    if isLastExp(exps): return eval(firstExp(exps), env)
    eval(firstExp(exps), env)
    return evalSequence(restExps(exps), env)


def evalAssignment(exp, env):
    setVariableValue(assignmentVariable(exp),
                     eval(assignmentValue(exp), env),
                     env)
    return Symbol("ok")


def evalDefinition(exp, env):
    defineVariable(definitionVariable(exp),
                   eval(definitionValue(exp), env),
                   env)
    return Symbol("ok")
                     

def isSelfEvaluating(exp):
    if isNumber(exp): return 1
    if isString(exp): return 1
    return 0


def isNumber(x):
    return type(x) in (types.IntType, types.FloatType, types.LongType)


def isString(x):
    return type(x) == types.StringType


def isVariable(exp):
    return isSymbol(exp)


def isQuoted(exp):
    return isTaggedList(exp, Symbol("quote"))


def textOfQuotation(exp):
    return exp[1]


def isTaggedList(exp, tag):
    if isList(exp):
        return exp[0] == tag
    return 0


def isList(x):
    return type(x) == types.ListType


######################################################################


def isAssignment(exp):
    """Assignments have the form (set! <var> <value>)."""
    return isTaggedList(exp, Symbol("set!"))


def assignmentVariable(exp):
    return exp[1]


def assignmentValue(exp):
    return exp[2]

######################################################################

def isDefinition(exp):
    return isTaggedList(exp, Symbol("define"))


def definitionVariable(exp):
    if isSymbol(exp[1]):
        return exp[1]
    return exp[1][0]


def definitionValue(exp):
    if isSymbol(exp[1]):
        return exp[2]
    return makeLambda(exp[1][1:], exp[2:])


######################################################################

def isLambda(exp):
    return isTaggedList(exp, Symbol("lambda"))


def lambdaParameters(exp):
    return exp[1]


def lambdaBody(exp):
    return exp[2:]


def makeLambda(parameters, body):
    return [Symbol("lambda"), parameters] + body


######################################################################


def isIf(exp):
    return isTaggedList(exp, Symbol("if"))


def ifPredicate(exp):
    return exp[1]


def ifConsequent(exp):
    return exp[2]


def ifAlternative(exp):
    if not isNull(exp[3:]):
        return exp[3]
    return false


def isNull(x):
    return x == []


def makeIf(predicate, consequent, alternative):
    return [Symbol("if"), predicate, consequent, alternative]


######################################################################

def isBegin(exp):
    return isTaggedList(exp, Symbol("begin"))


def beginActions(exp):
    return exp[1:]


def isLastExp(seq):
    return isNull(seq[1:])


def firstExp(seq):
    return seq[0]


def restExps(seq):
    return seq[1:]


def sequenceToExp(seq):
    if isNull(seq): return seq
    if isLastExp(seq): return firstExp(seq)
    return makeBegin(seq)


def makeBegin(seq):
    return [Symbol("begin")] + seq


######################################################################


def isApplication(exp):
    return isList(exp)


def operator(exp):
    return exp[0]


def operands(exp):
    return exp[1:]


def isNoOperands(ops):
    return isNull(ops)


def firstOperand(ops):
    return ops[0]


def restOperands(ops):
    return ops[1:]


######################################################################


def isCond(exp):
    return isTaggedList(exp, Symbol("cond"))


def condClauses(exp):
    return exp[1:]


def isCondElseClause(clause):
    return condPredicate(clause) == Symbol("else")


def condPredicate(clause):
    return clause[0]


def condActions(clause):
    return clause[1:]


def condToIf(exp):
    return expandClauses(condClauses(exp))


def expandClauses(clauses):
    if isNull(clauses):
        return false
    first = clauses[0]
    rest = clauses[1:]
    if isCondElseClause(first):
        if isNull(rest):
            return sequenceToExp(condActions(first))
        raise SchemeError, "else clause isn't last -- condToIf" + clauses
    return makeIf(condPredicate(first),
                  sequenceToExp(condActions(first)),
                  expandClauses(rest))
                                

def isTrue(x):
    return x != false


def isFalse(x):
    return x == false


######################################################################


def makeProcedure(parameters, body, env):
    return [Symbol("procedure"), parameters, body, env]

def isCompoundProcedure(p):
    return isTaggedList(p, Symbol("procedure"))

def procedureParameters(p):
    return p[1]

def procedureBody(p):
    return p[2]

def procedureEnvironment(p):
    return p[3]





######################################################################
"""Environment stuff.  Exercise 4.11 asks us to develop an alternative
environment implementation; I will try to use Python dictionaries
instead of name-value lists when I get the chance.

I deviate from the Scheme definitions on LOOKUP-VARIABLE-VALUE,
SET-VARIABLE-VALUE!, and DEFINE-VARIABLE! because of the awkwardness
of using a separate iterative SCAN function --- instead, I've used
Python's iterative control statements."""

def enclosingEnvironment(env):
    return env[1:]

def firstFrame(env):
    return env[0]

THE_EMPTY_ENVIRONMENT = []


def makeFrame(variables, values):
    """Note: here I diverge from SICP's implementation: instead of
    using a cons, I use a Python list.  The selectors frameVariables()
    and frameValues() reflect this."""
    return [variables, values]

def frameVariables(frame):
    return frame[0]

def frameValues(frame):
    return frame[1]

def addBindingToFrame(var, val, frame):
    frame[0] = [var] + frame[0]
    frame[1] = [val] + frame[1]


def extendEnvironment(vars, vals, base_env):
    if len(vars) == len(vals):
        return [makeFrame(vars, vals)] + base_env
    if len(vars) < len(vals):
        raise SchemeError, "Too many arguments supplied %s %s" % (vars, vals)
    raise SchemeError, "Too few arguments supplied %s %s" % (vars, vals)


def lookupVariableValue(var, env):
    while 1:
        if env == THE_EMPTY_ENVIRONMENT:
            raise SchemeError, "Unbound variable " + var
        frame = firstFrame(env)
        vars, vals = frameVariables(frame), frameValues(frame)
        for i in range(len(vars)):
            if vars[i] == var:
                return vals[i]
        env = enclosingEnvironment(env)


def setVariableValue(var, val, env):
    while 1:
        if env == THE_EMPTY_ENVIRONMENT:
            raise SchemeError, "Unbound variable -- SET! " + var
        frame = firstFrame(env)
        vars, values = frameVariables(frame), frameValues(frame)
        for i in range(len(vars)):
            if vars[i] == var:
                values[i] = val
                return
        env = enclosingEnvironment(env)


def defineVariable(var, val, env):
    frame = firstFrame(env)
    vars, values = frameVariables(frame), frameValues(frame)
    for i in range(len(vars)):
        if vars[i] == var:
            values[i] = val
            return
    addBindingToFrame(var, val, frame)


def setupEnvironment():
    initial_environment = extendEnvironment(PRIMITIVE_PROCEDURE_NAMES,
                                            PRIMITIVE_PROCEDURE_OBJECTS,
                                            THE_EMPTY_ENVIRONMENT)
    defineVariable(Symbol("true"), Symbol("true"), initial_environment)
    defineVariable(Symbol("false"), Symbol("false"), initial_environment)
    return initial_environment



######################################################################
"""Here's where we define the primitives of Scheme."""
def isPrimitiveProcedure(proc):
    return isTaggedList(proc, Symbol("primitive"))

def primitiveImplementation(proc): return proc[1]

def schemeCons(a, b):
    def dispatch(command):
        if command == 'car': return a
        if command == 'cdr': return b
        raise SchemeError, "CONS: invalid command '%s' on CONS pair" % command
    return dispatch

def schemeCar(pair): return pair("car")
def schemeCdr(pair): return pair("cdr")
def schemeAdd(a, b): return a + b
def schemeSubtract(a, b): return a - b
def schemeMultiply(a, b): return a * b
def schemeDivide(a, b): return a / b
def schemeRemainder(a, b): return a % b
def schemeQuotient(a, b): return int(a / b)
def schemeSqrt(a): return math.sqrt(a)
def schemeFloor(a): return math.floor(a)
def schemeNumericalEq(a, b):
    """Here, we need to make sure that we're returning the metacircular
    values of true or false."""
    if a == b: return true
    return false
def schemeDisplay(a):
    sys.stdout.write(str(a))   ## fixme
    return Symbol("ok")
def schemeNewline():
    sys.stdout.write("\n")
    return Symbol("ok")

PRIMITIVE_PROCEDURES = [
    [Symbol("car"), schemeCar],
    [Symbol("cdr"), schemeCdr],
    [Symbol("cons"), schemeCons],
    [Symbol("+"), schemeAdd],
    [Symbol("-"), schemeSubtract],
    [Symbol("*"), schemeMultiply],
    [Symbol("/"), schemeDivide],
    [Symbol("remainder"), schemeRemainder],
    [Symbol("quotient"), schemeQuotient],
    [Symbol("sqrt"), schemeSqrt],
    [Symbol("floor"), schemeFloor],
    [Symbol("="), schemeNumericalEq],
    [Symbol("display"), schemeDisplay],
    [Symbol("newline"), schemeNewline],
    ]
    
PRIMITIVE_PROCEDURE_NAMES = map(lambda x: x[0], PRIMITIVE_PROCEDURES)

PRIMITIVE_PROCEDURE_OBJECTS = map(lambda proc:
                                  [Symbol("primitive"), proc[1]],
                                  PRIMITIVE_PROCEDURES)

def applyPrimitiveProcedure(proc, args):
    return applyInUnderlyingPython(primitiveImplementation(proc),
                                   args)
######################################################################

def test():
    from parser import parse
    global_environment = setupEnvironment()
    commands = parse("""(
    (+ 3 4)
    (define pi 3.1415926)
    pi
    (* 4 2)
    (car (cdr (cons 3 (cons 4 5))))
    ((lambda (x) (* x x)) 42)
    ((lambda (x y z) (+ (* x y) z)) 3 4 5)
    (define square (lambda (x) (* x x)))
    (square 42)
    (define (hypotenuse a b) (sqrt (+ (square a) (square b))))
    (hypotenuse 3 4)
    (define (factorial x)
    (if (= x 0) 1
    (* x (factorial (- x 1)))))
    (factorial 0)
    (factorial 1)
    (factorial 2)
    (factorial 3)
    (begin
    (display 1)
    (display 2)
    (display 3)
    (newline)
    'is-this-working?)
    
    )""")
    for c in commands:
        pprint.pprint(c)
        print eval(c, global_environment)
        print
    return global_environment

if __name__ == '__main__': test()
